@@ -1,8 +1,5 @@
 package Gtk3;
-{
-  $Gtk3::VERSION = '0.014';
-}
-
+$Gtk3::VERSION = '0.018';
 use strict;
 use warnings;
 use Carp qw/croak/;
@@ -42,6 +39,7 @@ my @_GTK_FLATTEN_ARRAY_REF_RETURN_FOR = qw/
   Gtk3::Builder::get_objects
   Gtk3::CellLayout::get_cells
   Gtk3::Container::get_children
+  Gtk3::SizeGroup::get_widgets
   Gtk3::Stock::list_ids
   Gtk3::TreePath::get_indices
   Gtk3::TreeView::get_columns
@@ -973,22 +971,94 @@ sub Gtk3::MessageDialog::new {
   return $dialog;
 }
 
-# Gtk3::RadioMenuItem constructors.
+# Gtk3::RadioAction, Gtk3::RadioButton, Gtk3::RadioMenuItem and
+# Gtk3::RadioToolButton constructors.
 {
   no strict qw(refs);
-  foreach my $ctor (qw/new new_with_label new_with_mnemonic/) {
-    *{'Gtk3::RadioMenuItem::' . $ctor} = sub {
-      my ($class, $group_or_member, @rest) = @_;
-      my $real_ctor = $ctor;
-      {
-        local $@;
-        if (eval { $group_or_member->isa ('Gtk3::RadioMenuItem') }) {
+
+  my $group_converter = sub {
+    my ($ctor, $group_or_member, $package) = @_;
+    local $@;
+    # undef => []
+    if (!defined $group_or_member) {
+      return ($ctor, []);
+    }
+    # [] => []
+    elsif (eval { $#$group_or_member == -1 }) {
+      return ($ctor, []);
+    }
+    # [member1, ...] => member1
+    elsif (eval { $#$group_or_member >= 0 }) {
+      my $member = $group_or_member->[0];
+      if (defined $member) {
+        return ($ctor . '_from_widget', $member);
+      }
+      return ($ctor, []);
+    }
+    # member => member
+    elsif (eval { $group_or_member->isa ('Gtk3::' . $package) }) {
+      return ($ctor . '_from_widget', $group_or_member);
+    }
+    else {
+      croak ('Unhandled group or member argument encountered');
+    }
+  };
+
+  # Gtk3::RadioAction/Gtk3::RadioButton/Gtk3::RadioMenuItem/Gtk3::RadioToolButton
+  foreach my $package (qw/RadioAction RadioButton RadioMenuItem RadioToolButton/) {
+    *{'Gtk3::' . $package . '::set_group'} = sub {
+      my ($button, $group) = @_;
+      my $real_group = $group;
+      if (eval { $#$group >= 0 }) {
+        $real_group = $group->[0];
+      }
+      $button->set (group => $real_group);
+    };
+  }
+
+  # Gtk3::RadioButton/Gtk3::RadioMenuItem
+  foreach my $package (qw/RadioButton RadioMenuItem/) {
+    foreach my $ctor (qw/new new_with_label new_with_mnemonic/) {
+      # Avoid using the list-based API, as G:O:I does not support the memory
+      # ownership semantics.  Use the item-based API instead.
+      *{'Gtk3::' . $package . '::' . $ctor} = sub {
+        my ($class, $group_or_member, @rest) = @_;
+        my ($real_ctor, $real_group_or_member) =
+          $group_converter->($ctor, $group_or_member, $package);
+        return Glib::Object::Introspection->invoke (
+          $_GTK_BASENAME, $package, $real_ctor,
+          $class, $real_group_or_member, @rest);
+      };
+
+      # Work around <https://bugzilla.gnome.org/show_bug.cgi?id=679563>.
+      *{'Gtk3::' . $package . '::' . $ctor . '_from_widget'} = sub {
+        my ($class, $member, @rest) = @_;
+        my $real_ctor = $ctor;
+        my $real_group_or_member = $member;
+        if (!defined $member) {
+          $real_group_or_member = [];
+        } else {
           $real_ctor .= '_from_widget';
         }
-      }
+        return Glib::Object::Introspection->invoke (
+          $_GTK_BASENAME, $package, $real_ctor,
+          $class, $real_group_or_member, @rest);
+      };
+    }
+  }
+
+  # GtkRadioToolButton
+  foreach my $ctor (qw/new new_from_stock/) {
+    # Avoid using the list-based API, as G:O:I does not support the memory
+    # ownership semantics.  Use the item-based API instead.
+    *{'Gtk3::RadioToolButton::' . $ctor} = sub {
+      my ($class, $group_or_member, @rest) = @_;
+      my ($real_ctor, $real_group_or_member) =
+        $group_converter->($ctor, $group_or_member, 'RadioToolButton');
+      $real_ctor =~ s/_from_stock_from_/_with_stock_from_/; # you gotta be kidding me...
       return Glib::Object::Introspection->invoke (
-        $_GTK_BASENAME, 'RadioMenuItem', $real_ctor,
-        $class, $group_or_member, @rest);
+        $_GTK_BASENAME, 'RadioToolButton', $real_ctor,
+        $class, $real_group_or_member, @rest);
     };
   }
 }
@@ -1239,6 +1309,13 @@ sub Gtk3::VBox::new {
     $_GTK_BASENAME, 'VBox', 'new', $class, $homogeneous, $spacing);
 }
 
+sub Gtk3::Widget::render_icon {
+  my ($widget, $stock_id, $size, $detail) = @_;
+  Glib::Object::Introspection->invoke (
+    $_GTK_BASENAME, 'Widget', 'render_icon', $widget, $stock_id,
+    $_GTK_ICON_SIZE_NICK_TO_ID->($size), $detail);
+}
+
 sub Gtk3::Window::new {
   my ($class, $type) = @_;
   $type = 'toplevel' unless defined $type;
@@ -1386,6 +1463,57 @@ sub Gtk3::Gdk::Pixbuf::save_to_callback {
     $pixbuf, $save_func, $user_data, $type, $keys, $values);
 }
 
+# Pango
+
+sub Pango::Layout::set_text {
+  return Glib::Object::Introspection->invoke (
+    $_PANGO_BASENAME, 'Layout', 'set_text',
+    @_ == 3 ? @_ : (@_[0,1], -1)); # wants length in bytes
+}
+
+# - Fixes ------------------------------------------------------------------- #
+
+# Compatibility with perl 5.20 and non-dot locales.  Wrap all functions that
+# might end up calling setlocale() such that POSIX::setlocale() is also called
+# to ensure perl knows about the current locale.  See the discussion in
+# <https://rt.perl.org/Public/Bug/Display.html?id=121930>,
+# <https://rt.perl.org/Public/Bug/Display.html?id=121317>,
+# <https://rt.perl.org/Public/Bug/Display.html?id=120723>.
+if ($^V ge v5.20.0) {
+  require POSIX;
+  no strict 'refs';
+  no warnings 'redefine';
+
+  my $disable_setlocale = 0;
+  my $orig = \&Gtk3::disable_setlocale;
+  *{Gtk3::disable_setlocale} = sub {
+    $disable_setlocale = 1;
+    $orig->(@_);
+  };
+
+  # These two already have overrides.
+  foreach my $function (qw/Gtk3::init Gtk3::init_check/) {
+    my $orig = \&{$function};
+    *{$function} = sub {
+      if (!$disable_setlocale) {
+        POSIX::setlocale (POSIX::LC_ALL (), '');
+      }
+      $orig->(@_);
+    };
+  }
+
+  # These do not.
+  foreach my $function (qw/init_with_args parse_args/) {
+    *{'Gtk3::' . $function} = sub {
+      if (!$disable_setlocale) {
+        POSIX::setlocale (POSIX::LC_ALL (), '');
+      }
+      Glib::Object::Introspection->invoke (
+        $_GTK_BASENAME, undef, $function, @_);
+    };
+  }
+}
+
 # - Helpers ----------------------------------------------------------------- #
 
 sub _common_tree_model_new {
@@ -1538,6 +1666,8 @@ and must be returned.
 =item * Gtk3::Menu: The position callback passed to popup() does not receive x
 and y parameters anymore.
 
+=item * Gtk3::RadioAction: The constructor now follows the C API.
+
 =item * Gtk3::TreeModel: iter_next() is now a method that is modifying the iter
 directly, instead of returning a new one.  rows_reordered() and the
 "rows-reordered" signal are currently unusable.
